@@ -12,6 +12,7 @@ from pynndescent.optimal_transport import (
     network_simplex_core,
     total_cost,
     ProblemStatus,
+    sinkhorn_transport_plan,
 )
 
 _mock_identity = np.eye(2, dtype=np.float32)
@@ -22,7 +23,7 @@ FLOAT32_EPS = np.finfo(np.float32).eps
 FLOAT32_MAX = np.finfo(np.float32).max
 
 
-@numba.njit(fastmath=True, cache=True)
+@numba.njit(fastmath=True)
 def euclidean(x, y):
     r"""Standard euclidean distance.
 
@@ -66,7 +67,7 @@ def squared_euclidean(x, y):
     return result
 
 
-@numba.njit(fastmath=True, cache=True)
+@numba.njit(fastmath=True)
 def standardised_euclidean(x, y, sigma=_mock_ones):
     r"""Euclidean distance standardised against a vector of standard
     deviations per coordinate.
@@ -81,7 +82,7 @@ def standardised_euclidean(x, y, sigma=_mock_ones):
     return np.sqrt(result)
 
 
-@numba.njit(fastmath=True, cache=True)
+@numba.njit(fastmath=True)
 def manhattan(x, y):
     r"""Manhattan, taxicab, or l1 distance.
 
@@ -95,7 +96,7 @@ def manhattan(x, y):
     return result
 
 
-@numba.njit(fastmath=True, cache=True)
+@numba.njit(fastmath=True)
 def chebyshev(x, y):
     r"""Chebyshev or l-infinity distance.
 
@@ -109,7 +110,7 @@ def chebyshev(x, y):
     return result
 
 
-@numba.njit(fastmath=True, cache=True)
+@numba.njit(fastmath=True)
 def minkowski(x, y, p=2):
     r"""Minkowski distance.
 
@@ -128,7 +129,7 @@ def minkowski(x, y, p=2):
     return result ** (1.0 / p)
 
 
-@numba.njit(fastmath=True, cache=True)
+@numba.njit(fastmath=True)
 def weighted_minkowski(x, y, w=_mock_ones, p=2):
     r"""A weighted version of Minkowski distance.
 
@@ -141,12 +142,12 @@ def weighted_minkowski(x, y, w=_mock_ones, p=2):
     """
     result = 0.0
     for i in range(x.shape[0]):
-        result += (w[i] * np.abs(x[i] - y[i])) ** p
+        result += w[i] * np.abs(x[i] - y[i]) ** p
 
     return result ** (1.0 / p)
 
 
-@numba.njit(fastmath=True, cache=True)
+@numba.njit(fastmath=True)
 def mahalanobis(x, y, vinv=_mock_identity):
     result = 0.0
 
@@ -164,7 +165,7 @@ def mahalanobis(x, y, vinv=_mock_identity):
     return np.sqrt(result)
 
 
-@numba.njit(fastmath=True, cache=True)
+@numba.njit(fastmath=True)
 def hamming(x, y):
     result = 0.0
     for i in range(x.shape[0]):
@@ -174,7 +175,7 @@ def hamming(x, y):
     return float(result) / x.shape[0]
 
 
-@numba.njit(fastmath=True, cache=True)
+@numba.njit(fastmath=True)
 def canberra(x, y):
     result = 0.0
     for i in range(x.shape[0]):
@@ -185,7 +186,7 @@ def canberra(x, y):
     return result
 
 
-@numba.njit(fastmath=True, cache=True)
+@numba.njit(fastmath=True)
 def bray_curtis(x, y):
     numerator = 0.0
     denominator = 0.0
@@ -199,7 +200,7 @@ def bray_curtis(x, y):
         return 0.0
 
 
-@numba.njit(fastmath=True, cache=True)
+@numba.njit(fastmath=True)
 def jaccard(x, y):
     num_non_zero = 0.0
     num_equal = 0.0
@@ -255,7 +256,7 @@ def correct_alternative_jaccard(v):
     return 1.0 - pow(2.0, -v)
 
 
-@numba.njit(fastmath=True, cache=True)
+@numba.njit(fastmath=True)
 def matching(x, y):
     num_not_equal = 0.0
     for i in range(x.shape[0]):
@@ -266,7 +267,7 @@ def matching(x, y):
     return float(num_not_equal) / x.shape[0]
 
 
-@numba.njit(fastmath=True, cache=True)
+@numba.njit(fastmath=True)
 def dice(x, y):
     num_true_true = 0.0
     num_not_equal = 0.0
@@ -282,7 +283,7 @@ def dice(x, y):
         return num_not_equal / (2.0 * num_true_true + num_not_equal)
 
 
-@numba.njit(fastmath=True, cache=True)
+@numba.njit(fastmath=True)
 def kulsinski(x, y):
     num_true_true = 0.0
     num_not_equal = 0.0
@@ -300,7 +301,7 @@ def kulsinski(x, y):
         )
 
 
-@numba.njit(fastmath=True, cache=True)
+@numba.njit(fastmath=True)
 def rogers_tanimoto(x, y):
     num_not_equal = 0.0
     for i in range(x.shape[0]):
@@ -311,7 +312,7 @@ def rogers_tanimoto(x, y):
     return (2.0 * num_not_equal) / (x.shape[0] + num_not_equal)
 
 
-@numba.njit(fastmath=True, cache=True)
+@numba.njit(fastmath=True)
 def russellrao(x, y):
     num_true_true = 0.0
     for i in range(x.shape[0]):
@@ -325,7 +326,7 @@ def russellrao(x, y):
         return float(x.shape[0] - num_true_true) / (x.shape[0])
 
 
-@numba.njit(fastmath=True, cache=True)
+@numba.njit(fastmath=True)
 def sokal_michener(x, y):
     num_not_equal = 0.0
     for i in range(x.shape[0]):
@@ -336,7 +337,7 @@ def sokal_michener(x, y):
     return (2.0 * num_not_equal) / (x.shape[0] + num_not_equal)
 
 
-@numba.njit(fastmath=True, cache=True)
+@numba.njit(fastmath=True)
 def sokal_sneath(x, y):
     num_true_true = 0.0
     num_not_equal = 0.0
@@ -352,7 +353,7 @@ def sokal_sneath(x, y):
         return num_not_equal / (0.5 * num_true_true + num_not_equal)
 
 
-@numba.njit(fastmath=True, cache=True)
+@numba.njit(fastmath=True)
 def haversine(x, y):
     if x.shape[0] != 2:
         raise ValueError("haversine is only defined for 2 dimensional graph_data")
@@ -362,7 +363,7 @@ def haversine(x, y):
     return 2.0 * np.arcsin(result)
 
 
-@numba.njit(fastmath=True, cache=True)
+@numba.njit(fastmath=True)
 def yule(x, y):
     num_true_true = 0.0
     num_true_false = 0.0
@@ -384,7 +385,7 @@ def yule(x, y):
         )
 
 
-@numba.njit(fastmath=True, cache=True)
+@numba.njit(fastmath=True)
 def cosine(x, y):
     result = 0.0
     norm_x = 0.0
@@ -546,7 +547,7 @@ def true_angular_from_alt_cosine(d):
     return 1.0 - (np.arccos(pow(2.0, -d)) / np.pi)
 
 
-@numba.njit(fastmath=True, cache=True)
+@numba.njit(fastmath=True)
 def correlation(x, y):
     mu_x = 0.0
     mu_y = 0.0
@@ -728,9 +729,7 @@ def kantorovich(x, y, cost=_dummy_cost, max_iter=100000):
     sub_cost = cost[row_mask, :][:, col_mask]
 
     node_arc_data, spanning_tree, graph = allocate_graph_structures(
-        a.shape[0],
-        b.shape[0],
-        False,
+        a.shape[0], b.shape[0], False
     )
     initialize_supply(a, -b, graph, node_arc_data.supply)
     initialize_cost(sub_cost, graph, node_arc_data.cost)
@@ -740,12 +739,7 @@ def kantorovich(x, y, cost=_dummy_cost, max_iter=100000):
         raise ValueError(
             "Kantorovich distance inputs must be valid probability distributions."
         )
-    solve_status = network_simplex_core(
-        node_arc_data,
-        spanning_tree,
-        graph,
-        max_iter,
-    )
+    solve_status = network_simplex_core(node_arc_data, spanning_tree, graph, max_iter)
     # if solve_status == ProblemStatus.MAX_ITER_REACHED:
     #     print("WARNING: RESULT MIGHT BE INACCURATE\nMax number of iteration reached!")
     if solve_status == ProblemStatus.INFEASIBLE:
@@ -761,6 +755,146 @@ def kantorovich(x, y, cost=_dummy_cost, max_iter=100000):
     return result
 
 
+@numba.njit(fastmath=True)
+def sinkhorn(x, y, cost=_dummy_cost, regularization=1.0):
+    row_mask = x != 0
+    col_mask = y != 0
+
+    a = x[row_mask].astype(np.float64)
+    b = y[col_mask].astype(np.float64)
+
+    a_sum = a.sum()
+    b_sum = b.sum()
+
+    a /= a_sum
+    b /= b_sum
+
+    sub_cost = cost[row_mask, :][:, col_mask]
+
+    transport_plan = sinkhorn_transport_plan(
+        x, y, cost=sub_cost, regularization=regularization
+    )
+    dim_i = transport_plan.shape[0]
+    dim_j = transport_plan.shape[1]
+    result = 0.0
+    for i in range(dim_i):
+        for j in range(dim_j):
+            result += transport_plan[i, j] * cost[i, j]
+
+    return result
+
+
+@numba.njit()
+def jensen_shannon_divergence(x, y):
+    result = 0.0
+    l1_norm_x = 0.0
+    l1_norm_y = 0.0
+    dim = x.shape[0]
+
+    for i in range(dim):
+        l1_norm_x += x[i]
+        l1_norm_y += y[i]
+
+    l1_norm_x += FLOAT32_EPS * dim
+    l1_norm_y += FLOAT32_EPS * dim
+
+    pdf_x = (x + FLOAT32_EPS) / l1_norm_x
+    pdf_y = (y + FLOAT32_EPS) / l1_norm_y
+    m = 0.5 * (pdf_x + pdf_y)
+
+    for i in range(dim):
+        result += 0.5 * (
+            pdf_x[i] * np.log(pdf_x[i] / m[i]) + pdf_y[i] * np.log(pdf_y[i] / m[i])
+        )
+
+    return result
+
+
+@numba.njit()
+def wasserstein_1d(x, y, p=1):
+    x_sum = 0.0
+    y_sum = 0.0
+    for i in range(x.shape[0]):
+        x_sum += x[i]
+        y_sum += y[i]
+
+    x_cdf = x / x_sum
+    y_cdf = y / y_sum
+
+    for i in range(1, x_cdf.shape[0]):
+        x_cdf[i] += x_cdf[i - 1]
+        y_cdf[i] += y_cdf[i - 1]
+
+    return minkowski(x_cdf, y_cdf, p)
+
+
+@numba.njit()
+def circular_kantorovich(x, y, p=1):
+    x_sum = 0.0
+    y_sum = 0.0
+    for i in range(x.shape[0]):
+        x_sum += x[i]
+        y_sum += y[i]
+
+    x_cdf = x / x_sum
+    y_cdf = y / y_sum
+
+    for i in range(1, x_cdf.shape[0]):
+        x_cdf[i] += x_cdf[i - 1]
+        y_cdf[i] += y_cdf[i - 1]
+
+    mu = np.median((x_cdf - y_cdf) ** p)
+
+    # Now we just want minkowski distance on the CDFs shifted by mu
+    result = 0.0
+    if p > 2:
+        for i in range(x_cdf.shape[0]):
+            result += np.abs(x_cdf[i] - y_cdf[i] - mu) ** p
+
+        return result ** (1.0 / p)
+
+    elif p == 2:
+        for i in range(x_cdf.shape[0]):
+            val = x_cdf[i] - y_cdf[i] - mu
+            result += val * val
+
+        return np.sqrt(result)
+
+    elif p == 1:
+        for i in range(x_cdf.shape[0]):
+            result += np.abs(x_cdf[i] - y_cdf[i] - mu)
+
+        return result
+
+    else:
+        raise ValueError("Invalid p supplied to Kantorvich distance")
+
+
+@numba.njit()
+def symmetric_kl_divergence(x, y):
+    result = 0.0
+    l1_norm_x = 0.0
+    l1_norm_y = 0.0
+    dim = x.shape[0]
+
+    for i in range(dim):
+        l1_norm_x += x[i]
+        l1_norm_y += y[i]
+
+    l1_norm_x += FLOAT32_EPS * dim
+    l1_norm_y += FLOAT32_EPS * dim
+
+    pdf_x = (x + FLOAT32_EPS) / l1_norm_x
+    pdf_y = (y + FLOAT32_EPS) / l1_norm_y
+
+    for i in range(dim):
+        result += pdf_x[i] * np.log(pdf_x[i] / pdf_y[i]) + pdf_y[i] * np.log(
+            pdf_y[i] / pdf_x[i]
+        )
+
+    return result
+
+
 named_distances = {
     # general minkowski distances
     "euclidean": euclidean,
@@ -785,14 +919,27 @@ named_distances = {
     "cosine": cosine,
     "dot": dot,
     "correlation": correlation,
-    "hellinger": hellinger,
     "haversine": haversine,
     "braycurtis": bray_curtis,
     "spearmanr": spearmanr,
-    "kantorovich": kantorovich,
-    "wasserstein": kantorovich,
     "tsss": tsss,
     "true_angular": true_angular,
+    # Distribution distances
+    "hellinger": hellinger,
+    "kantorovich": kantorovich,
+    "wasserstein": kantorovich,
+    "wasserstein_1d": wasserstein_1d,
+    "wasserstein-1d": wasserstein_1d,
+    "kantorovich-1d": wasserstein_1d,
+    "kantorovich_1d": wasserstein_1d,
+    "circular_kantorovich": circular_kantorovich,
+    "circular_wasserstein": circular_kantorovich,
+    "sinkhorn": sinkhorn,
+    "jensen-shannon": jensen_shannon_divergence,
+    "jensen_shannon": jensen_shannon_divergence,
+    "symmetric-kl": symmetric_kl_divergence,
+    "symmetric_kl": symmetric_kl_divergence,
+    "symmetric_kullback_liebler": symmetric_kl_divergence,
     # Binary distances
     "hamming": hamming,
     "jaccard": jaccard,
