@@ -6,11 +6,9 @@ from warnings import warn
 import locale
 import numpy as np
 import numba
-from numba.core import types
-from numba.experimental import structref
 import scipy.sparse
 
-from pynndescent.sparse import sparse_mul, sparse_diff, sparse_sum, arr_intersect
+from pynndescent.sparse import sparse_mul, sparse_diff, sparse_sum, arr_intersect, sparse_dot_product
 from pynndescent.utils import tau_rand_int, norm
 import joblib
 
@@ -548,13 +546,11 @@ def make_euclidean_tree(
         offsets.append(offset)
         children.append((np.int32(left_node_num), np.int32(right_node_num)))
         point_indices.append(np.array([-1], dtype=np.int32))
-        # print("Made a node in tree with", len(point_indices), "nodes")
     else:
         hyperplanes.append(np.array([-1.0], dtype=np.float32))
         offsets.append(-np.inf)
         children.append((np.int32(-1), np.int32(-1)))
         point_indices.append(indices)
-        # print("Made a leaf in tree with", len(point_indices), "nodes")
 
     return
 
@@ -795,9 +791,7 @@ def make_dense_tree(data, rng_state, leaf_size=30, angular=False):
             leaf_size,
         )
 
-    # print("Completed a tree")
     result = FlatTree(hyperplanes, offsets, children, point_indices, leaf_size)
-    # print("Tree type is:", numba.typeof(result))
     return result
 
 
@@ -856,6 +850,7 @@ def make_sparse_tree(inds, indptr, spdata, rng_state, leaf_size=30, angular=Fals
         "dim": numba.types.intp,
         "d": numba.types.uint16,
     },
+    cache=True,
 )
 def select_side(hyperplane, offset, point, rng_state):
     margin = offset
@@ -888,6 +883,7 @@ def select_side(hyperplane, offset, point, rng_state):
         ),
     ],
     locals={"node": numba.types.uint32, "side": numba.types.boolean},
+    cache=True,
 )
 def search_flat_tree(point, hyperplanes, offsets, children, indices, rng_state):
     node = 0
@@ -901,7 +897,7 @@ def search_flat_tree(point, hyperplanes, offsets, children, indices, rng_state):
     return indices[-children[node, 0] : -children[node, 1]]
 
 
-@numba.njit(fastmath=True)
+@numba.njit(fastmath=True, cache=True)
 def sparse_select_side(hyperplane, offset, point_inds, point_data, rng_state):
     margin = offset
 
@@ -912,10 +908,7 @@ def sparse_select_side(hyperplane, offset, point_inds, point_data, rng_state):
     hyperplane_inds = hyperplane[0, :hyperplane_size].astype(np.int32)
     hyperplane_data = hyperplane[1, :hyperplane_size]
 
-    _, aux_data = sparse_mul(hyperplane_inds, hyperplane_data, point_inds, point_data)
-
-    for val in aux_data:
-        margin += val
+    margin += sparse_dot_product(hyperplane_inds, hyperplane_data, point_inds, point_data)
 
     if abs(margin) < EPS:
         side = tau_rand_int(rng_state) % 2
@@ -929,7 +922,7 @@ def sparse_select_side(hyperplane, offset, point_inds, point_data, rng_state):
         return 1
 
 
-@numba.njit(locals={"node": numba.types.uint32})
+@numba.njit(locals={"node": numba.types.uint32}, cache=True)
 def search_sparse_flat_tree(
     point_inds, point_data, hyperplanes, offsets, children, indices, rng_state
 ):
@@ -975,7 +968,7 @@ def make_forest(
     # print(ts(), "Started forest construction")
     result = []
     if leaf_size is None:
-        leaf_size = max(10, n_neighbors)
+        leaf_size = max(10, np.int32(n_neighbors))
     if n_jobs is None:
         n_jobs = -1
 
@@ -1010,7 +1003,7 @@ def make_forest(
     return tuple(result)
 
 
-@numba.njit(nogil=True)
+@numba.njit(nogil=True, cache=True)
 def get_leaves_from_tree(tree):
     n_leaves = 0
     for i in range(len(tree.children)):
@@ -1124,7 +1117,7 @@ def recursive_convert_sparse(
         return node_num, leaf_start
 
 
-@numba.njit()
+@numba.njit(cache=True)
 def num_nodes_and_leaves(tree):
     n_nodes = 0
     n_leaves = 0
@@ -1138,7 +1131,7 @@ def num_nodes_and_leaves(tree):
     return n_nodes, n_leaves
 
 
-@numba.njit()
+@numba.njit(cache=True)
 def dense_hyperplane_dim(hyperplanes):
     for i in range(len(hyperplanes)):
         if hyperplanes[i].shape[0] > 1:
@@ -1147,7 +1140,7 @@ def dense_hyperplane_dim(hyperplanes):
     raise ValueError("No hyperplanes of adequate size were found!")
 
 
-@numba.njit()
+@numba.njit(cache=True)
 def sparse_hyperplane_dim(hyperplanes):
     max_dim = 0
     for i in range(len(hyperplanes)):
@@ -1226,6 +1219,7 @@ def renumbaify_tree(tree):
         "result": numba.float32,
         "i": numba.uint32,
     },
+    cache=False,
 )
 def score_tree(tree, neighbor_indices, data, rng_state):
     result = 0.0
@@ -1243,11 +1237,11 @@ def score_tree(tree, neighbor_indices, data, rng_state):
     return result / numba.float32(neighbor_indices.shape[0])
 
 
-@numba.njit(nogil=True, parallel=True, locals={"node": numba.int32})
+@numba.njit(nogil=True, locals={"node": numba.int32}, cache=False)
 def score_linked_tree(tree, neighbor_indices):
     result = 0.0
     n_nodes = len(tree.children)
-    for i in numba.prange(n_nodes):
+    for i in range(n_nodes):
         node = numba.int32(i)
         left_child = tree.children[node][0]
         right_child = tree.children[node][1]
