@@ -8,7 +8,11 @@ import numpy as np
 import numba
 
 from pynndescent.utils import norm, tau_rand
-from pynndescent.distances import kantorovich
+from pynndescent.distances import (
+    kantorovich,
+    jensen_shannon_divergence,
+    symmetric_kl_divergence,
+)
 
 locale.setlocale(locale.LC_NUMERIC, "C")
 
@@ -16,14 +20,14 @@ FLOAT32_EPS = np.finfo(np.float32).eps
 FLOAT32_MAX = np.finfo(np.float32).max
 
 # Just reproduce a simpler version of numpy isclose (not numba supported yet)
-@numba.njit()
+@numba.njit(cache=True)
 def isclose(a, b, rtol=1.0e-5, atol=1.0e-8):
     diff = np.abs(a - b)
     return diff <= (atol + rtol * np.abs(b))
 
 
 # Just reproduce a simpler version of numpy unique (not numba supported yet)
-@numba.njit()
+@numba.njit(cache=True)
 def arr_unique(arr):
     aux = np.sort(arr)
     flag = np.concatenate((np.ones(1, dtype=np.bool_), aux[1:] != aux[:-1]))
@@ -31,7 +35,7 @@ def arr_unique(arr):
 
 
 # Just reproduce a simpler version of numpy union1d (not numba supported yet)
-@numba.njit()
+@numba.njit(cache=True)
 def arr_union(ar1, ar2):
     if ar1.shape[0] == 0:
         return ar2
@@ -43,12 +47,66 @@ def arr_union(ar1, ar2):
 
 # Just reproduce a simpler version of numpy intersect1d (not numba supported
 # yet)
-@numba.njit()
+@numba.njit(cache=True)
 def arr_intersect(ar1, ar2):
     aux = np.concatenate((ar1, ar2))
     aux.sort()
     return aux[:-1][aux[1:] == aux[:-1]]
 
+# Some things require size of intersection; do this quickly; assume sorted arrays for speed
+@numba.njit(
+    [
+        "i4(i4[:],i4[:])",
+        numba.types.int32(
+            numba.types.Array(numba.types.int32, 1, "C", readonly=True),
+            numba.types.Array(numba.types.int32, 1, "C", readonly=True),
+        ),
+    ],
+    locals={
+        "i1": numba.uint16,
+        "i2": numba.uint16,
+    }
+)
+def fast_intersection_size(ar1, ar2):
+    if ar1.shape[0] == 0 or ar2.shape[0] == 0:
+        return 0
+
+    # NOTE: We assume arrays are sorted; if they are not this will break
+    i1 = 0
+    i2 = 0
+    limit1 = ar1.shape[0] - 1
+    limit2 = ar2.shape[0] - 1
+    j1 = ar1[i1]
+    j2 = ar2[i2]
+
+    result = 0
+
+    while True:
+        if j1 == j2:
+            result += 1
+            if i1 < limit1:
+                i1 += 1
+                j1 = ar1[i1]
+            else:
+                break
+
+            if i2 < limit2:
+                i2 += 1
+                j2 = ar2[i2]
+            else:
+                break
+
+        elif j1 < j2 and i1 < limit1:
+            i1 += 1
+            j1 = ar1[i1]
+        elif j2 < j1 and i2 < limit2:
+            i2 += 1
+            j2 = ar2[i2]
+        else:
+            break
+
+    return result
+
 
 @numba.njit(
     [
@@ -62,7 +120,7 @@ def arr_intersect(ar1, ar2):
             numba.types.Array(numba.types.float32, 1, "C", readonly=True),
             numba.types.Array(numba.types.int32, 1, "C", readonly=True),
             numba.types.Array(numba.types.float32, 1, "C", readonly=True),
-        ),
+        )
     ],
     fastmath=True,
     locals={
@@ -74,6 +132,7 @@ def arr_intersect(ar1, ar2):
         "j1": numba.types.int32,
         "j2": numba.types.int32,
     },
+    cache=True,
 )
 def sparse_sum(ind1, data1, ind2, data2):
     result_size = ind1.shape[0] + ind2.shape[0]
@@ -138,11 +197,10 @@ def sparse_sum(ind1, data1, ind2, data2):
     return result_ind, result_data
 
 
-@numba.njit()
+@numba.njit(cache=True)
 def sparse_diff(ind1, data1, ind2, data2):
     return sparse_sum(ind1, data1, ind2, -data2)
 
-
 @numba.njit(
     [
         # "Tuple((i4[::1],f4[::1]))(i4[::1],f4[::1],i4[::1],f4[::1])",
@@ -156,7 +214,7 @@ def sparse_diff(ind1, data1, ind2, data2):
             numba.types.Array(numba.types.float32, 1, "C", readonly=True),
             numba.types.Array(numba.types.int32, 1, "C", readonly=True),
             numba.types.Array(numba.types.float32, 1, "C", readonly=True),
-        ),
+        )
     ],
     fastmath=True,
     locals={
@@ -166,6 +224,7 @@ def sparse_diff(ind1, data1, ind2, data2):
         "j1": numba.types.int32,
         "j2": numba.types.int32,
     },
+    cache=True,
 )
 def sparse_mul(ind1, data1, ind2, data2):
     result_ind = numba.typed.List.empty_list(numba.types.int32)
@@ -193,8 +252,124 @@ def sparse_mul(ind1, data1, ind2, data2):
 
     return result_ind, result_data
 
+@numba.njit(
+    [
+        # "Tuple((i4[::1],f4[::1]))(i4[::1],f4[::1],i4[::1],f4[::1])",
+        numba.types.float32(
+            numba.types.Array(numba.types.int32, 1, "C", readonly=True),
+            numba.types.Array(numba.types.float32, 1, "C", readonly=True),
+            numba.types.Array(numba.types.int32, 1, "C", readonly=True),
+            numba.types.Array(numba.types.float32, 1, "C", readonly=True),
+        )
+    ],
+    fastmath=True,
+    locals={
+        "result": numba.types.float32,
+        "val": numba.types.float32,
+        "i1": numba.types.uint16,
+        "i2": numba.types.uint16,
+        "j1": numba.types.int32,
+        "j2": numba.types.int32,
+    },
+    cache=True,
+)
+def sparse_dot_product(ind1, data1, ind2, data2):
+    dim1 = ind1.shape[0]
+    dim2 = ind2.shape[0]
 
+    result = 0.0
+
+    i1 = 0
+    i2 = 0
+    j1 = ind1[i1]
+    j2 = ind2[i2]
+
+    # pass through both index lists
+    while True:
+        if j1 == j2:
+            val = data1[i1] * data2[i2]
+            result += val
+            i1 += 1
+            if i1 >= dim1:
+                return result
+            j1 = ind1[i1]
+            i2 += 1
+            if i2 >= dim2:
+                return result
+            j2 = ind2[i2]
+        elif j1 < j2:
+            i1 += 1
+            if i1 >= dim1:
+                return result
+            j1 = ind1[i1]
+        else:
+            i2 += 1
+            if i2 >= dim2:
+                return result
+            j2 = ind2[i2]
+
+    return result # unreachable
+
+# Return dense vectors supported on the union of the non-zero valued indices
 @numba.njit()
+def dense_union(ind1, data1, ind2, data2):
+    result_ind = arr_union(ind1, ind2)
+    result_data1 = np.zeros(result_ind.shape[0], dtype=np.float32)
+    result_data2 = np.zeros(result_ind.shape[0], dtype=np.float32)
+
+    i1 = 0
+    i2 = 0
+    nnz = 0
+
+    # pass through both index lists
+    while i1 < ind1.shape[0] and i2 < ind2.shape[0]:
+        j1 = ind1[i1]
+        j2 = ind2[i2]
+
+        if j1 == j2:
+            val = data1[i1] + data2[i2]
+            if val != 0:
+                result_data1[nnz] = data1[i1]
+                result_data2[nnz] = data2[i2]
+                nnz += 1
+            i1 += 1
+            i2 += 1
+        elif j1 < j2:
+            val = data1[i1]
+            if val != 0:
+                result_data1[nnz] = data1[i1]
+                nnz += 1
+            i1 += 1
+        else:
+            val = data2[i2]
+            if val != 0:
+                result_data2[nnz] = data2[i2]
+                nnz += 1
+            i2 += 1
+
+    # pass over the tails
+    while i1 < ind1.shape[0]:
+        val = data1[i1]
+        if val != 0:
+            result_data1[nnz] = data1[i1]
+            nnz += 1
+        i1 += 1
+
+    while i2 < ind2.shape[0]:
+        val = data2[i2]
+        if val != 0:
+            result_data2[nnz] = data2[i2]
+            nnz += 1
+        i2 += 1
+
+    # truncate to the correct length in case there were zeros
+    result_data1 = result_data1[:nnz]
+    result_data2 = result_data2[:nnz]
+
+    return result_data1, result_data2
+
+
+@numba.njit(fastmath=True)
 def sparse_euclidean(ind1, data1, ind2, data2):
     _, aux_data = sparse_diff(ind1, data1, ind2, data2)
     result = 0.0
@@ -315,8 +490,8 @@ def sparse_bray_curtis(ind1, data1, ind2, data2):  # pragma: no cover
 
 @numba.njit()
 def sparse_jaccard(ind1, data1, ind2, data2):
-    num_non_zero = arr_union(ind1, ind2).shape[0]
-    num_equal = arr_intersect(ind1, ind2).shape[0]
+    num_equal = fast_intersection_size(ind1, ind2)
+    num_non_zero = ind1.shape[0] + ind2.shape[0] - num_equal
 
     if num_non_zero == 0:
         return 0.0
@@ -335,30 +510,31 @@ def sparse_jaccard(ind1, data1, ind2, data2):
         ),
     ],
     fastmath=True,
-    locals={
-        "num_non_zero": numba.types.intp,
-        "num_equal": numba.types.intp,
-    },
+    locals={"num_non_zero": numba.types.intp, "num_equal": numba.types.intp},
 )
 def sparse_alternative_jaccard(ind1, data1, ind2, data2):
-    num_non_zero = arr_union(ind1, ind2).shape[0]
-    num_equal = arr_intersect(ind1, ind2).shape[0]
+    num_equal = fast_intersection_size(ind1, ind2)
+    num_non_zero = ind1.shape[0] + ind2.shape[0] - num_equal
 
     if num_non_zero == 0:
         return 0.0
+    elif num_equal == 0:
+        return FLOAT32_MAX
     else:
         return -np.log2(num_equal / num_non_zero)
+        # return (num_non_zero - num_equal) / num_equal
 
 
 @numba.vectorize(fastmath=True)
 def correct_alternative_jaccard(v):
     return 1.0 - pow(2.0, -v)
+    # return v / (v + 1)
 
 
 @numba.njit()
 def sparse_matching(ind1, data1, ind2, data2, n_features):
-    num_true_true = arr_intersect(ind1, ind2).shape[0]
-    num_non_zero = arr_union(ind1, ind2).shape[0]
+    num_true_true = fast_intersection_size(ind1, ind2)
+    num_non_zero = ind1.shape[0] + ind2.shape[0] - num_true_true
     num_not_equal = num_non_zero - num_true_true
 
     return float(num_not_equal) / n_features
@@ -366,8 +542,8 @@ def sparse_matching(ind1, data1, ind2, data2, n_features):
 
 @numba.njit()
 def sparse_dice(ind1, data1, ind2, data2):
-    num_true_true = arr_intersect(ind1, ind2).shape[0]
-    num_non_zero = arr_union(ind1, ind2).shape[0]
+    num_true_true = fast_intersection_size(ind1, ind2)
+    num_non_zero = ind1.shape[0] + ind2.shape[0] - num_true_true
     num_not_equal = num_non_zero - num_true_true
 
     if num_not_equal == 0.0:
@@ -378,8 +554,8 @@ def sparse_dice(ind1, data1, ind2, data2):
 
 @numba.njit()
 def sparse_kulsinski(ind1, data1, ind2, data2, n_features):
-    num_true_true = arr_intersect(ind1, ind2).shape[0]
-    num_non_zero = arr_union(ind1, ind2).shape[0]
+    num_true_true = fast_intersection_size(ind1, ind2)
+    num_non_zero = ind1.shape[0] + ind2.shape[0] - num_true_true
     num_not_equal = num_non_zero - num_true_true
 
     if num_not_equal == 0:
@@ -392,8 +568,8 @@ def sparse_kulsinski(ind1, data1, ind2, data2, n_features):
 
 @numba.njit()
 def sparse_rogers_tanimoto(ind1, data1, ind2, data2, n_features):
-    num_true_true = arr_intersect(ind1, ind2).shape[0]
-    num_non_zero = arr_union(ind1, ind2).shape[0]
+    num_true_true = fast_intersection_size(ind1, ind2)
+    num_non_zero = ind1.shape[0] + ind2.shape[0] - num_true_true
     num_not_equal = num_non_zero - num_true_true
 
     return (2.0 * num_not_equal) / (n_features + num_not_equal)
@@ -404,7 +580,7 @@ def sparse_russellrao(ind1, data1, ind2, data2, n_features):
     if ind1.shape[0] == ind2.shape[0] and np.all(ind1 == ind2):
         return 0.0
 
-    num_true_true = arr_intersect(ind1, ind2).shape[0]
+    num_true_true = fast_intersection_size(ind1, ind2)
 
     if num_true_true == np.sum(data1 != 0) and num_true_true == np.sum(data2 != 0):
         return 0.0
@@ -414,8 +590,8 @@ def sparse_russellrao(ind1, data1, ind2, data2, n_features):
 
 @numba.njit()
 def sparse_sokal_michener(ind1, data1, ind2, data2, n_features):
-    num_true_true = arr_intersect(ind1, ind2).shape[0]
-    num_non_zero = arr_union(ind1, ind2).shape[0]
+    num_true_true = fast_intersection_size(ind1, ind2)
+    num_non_zero = ind1.shape[0] + ind2.shape[0] - num_true_true
     num_not_equal = num_non_zero - num_true_true
 
     return (2.0 * num_not_equal) / (n_features + num_not_equal)
@@ -423,8 +599,8 @@ def sparse_sokal_michener(ind1, data1, ind2, data2, n_features):
 
 @numba.njit()
 def sparse_sokal_sneath(ind1, data1, ind2, data2):
-    num_true_true = arr_intersect(ind1, ind2).shape[0]
-    num_non_zero = arr_union(ind1, ind2).shape[0]
+    num_true_true = fast_intersection_size(ind1, ind2)
+    num_non_zero = ind1.shape[0] + ind2.shape[0] - num_true_true
     num_not_equal = num_non_zero - num_true_true
 
     if num_not_equal == 0.0:
@@ -491,11 +667,7 @@ def sparse_correct_alternative_cosine(d):
 
 @numba.njit()
 def sparse_dot(ind1, data1, ind2, data2):
-    _, aux_data = sparse_mul(ind1, data1, ind2, data2)
-    result = 0.0
-
-    for val in aux_data:
-        result += val
+    result = sparse_dot_product(ind1, data1, ind2, data2)
 
     return 1.0 - result
 
@@ -505,16 +677,10 @@ def sparse_dot(ind1, data1, ind2, data2):
     fastmath=True,
     locals={
         "result": numba.types.float32,
-        "dim": numba.types.intp,
-        "i": numba.types.uint16,
     },
 )
 def sparse_alternative_dot(ind1, data1, ind2, data2):
-    _, aux_data = sparse_mul(ind1, data1, ind2, data2)
-    result = 0.0
-    dim = len(aux_data)
-    for i in range(dim):
-        result += aux_data[i]
+    result = sparse_dot_product(ind1, data1, ind2, data2)
 
     if result <= 0.0:
         return FLOAT32_MAX
@@ -696,7 +862,83 @@ def sparse_kantorovich(ind1, data1, ind2, data2, ground_metric=dummy_ground_metr
     return kantorovich(data1, data2, cost_matrix)
 
 
-@numba.njit(parallel=True)
+@numba.njit()
+def sparse_wasserstein_1d(ind1, data1, ind2, data2, p=1):
+    result = 0.0
+    old_ind = 0
+    delta = 0.0
+    i1 = 0
+    i2 = 0
+    cdf1 = 0.0
+    cdf2 = 0.0
+    l1_norm_x = np.sum(data1)
+    l1_norm_y = np.sum(data2)
+
+    norm = lambda x, p: np.power(np.abs(x), p)
+
+    # pass through both index lists
+    while i1 < ind1.shape[0] and i2 < ind2.shape[0]:
+        j1 = ind1[i1]
+        j2 = ind2[i2]
+
+        if j1 == j2:
+            result += delta * (j1 - old_ind)
+            cdf1 += data1[i1] / l1_norm_x
+            cdf2 += data2[i2] / l1_norm_y
+            delta = norm(cdf1 - cdf2, p)
+            old_ind = j1
+            i1 += 1
+            i2 += 1
+        elif j1 < j2:
+            result += delta * (j1 - old_ind)
+            cdf1 += data1[i1] / l1_norm_x
+            delta = norm(cdf1 - cdf2, p)
+            old_ind = j1
+            i1 += 1
+        else:
+            result += delta * (j2 - old_ind)
+            cdf2 += data2[i2] / l1_norm_y
+            delta = norm(cdf1 - cdf2, p)
+            old_ind = j2
+            i2 += 1
+            # pass over the tails
+    while i1 < ind1.shape[0]:
+        j1 = ind1[i1]
+        result += delta * (j1 - old_ind)
+        cdf1 += data1[i1] / l1_norm_x
+        delta = norm(cdf1 - cdf2, p)
+        old_ind = j1
+        i1 += 1
+
+    while i2 < ind2.shape[0]:
+        j2 = ind2[i2]
+        result += delta * (j2 - old_ind)
+        cdf2 += data2[i2] / l1_norm_y
+        delta = norm(cdf1 - cdf2, p)
+        old_ind = j2
+        i2 += 1
+
+    return np.power(result, 1.0 / p)
+
+
+# Because of the EPS values and the need to normalize after adding them (and then average those for jensen_shannon)
+# it seems like we might as well just take the dense union (dense vectors supported on the union of indices)
+# and call the dense distance functions
+
+
+@numba.njit()
+def sparse_jensen_shannon_divergence(ind1, data1, ind2, data2):
+    dense_data1, dense_data2 = dense_union(ind1, data1, ind2, data2)
+    return jensen_shannon_divergence(dense_data1, dense_data2)
+
+
+@numba.njit()
+def sparse_symmetric_kl_divergence(ind1, data1, ind2, data2):
+    dense_data1, dense_data2 = dense_union(ind1, data1, ind2, data2)
+    return symmetric_kl_divergence(dense_data1, dense_data2)
+
+
+@numba.njit(parallel=True, cache=False)
 def diversify(
     indices,
     distances,
@@ -751,7 +993,7 @@ def diversify(
     return indices, distances
 
 
-@numba.njit(parallel=True)
+@numba.njit(parallel=True, cache=False)
 def diversify_csr(
     graph_indptr,
     graph_indices,
@@ -821,8 +1063,6 @@ sparse_named_distances = {
     "minkowski": sparse_minkowski,
     # Other distances
     "canberra": sparse_canberra,
-    "kantorovich": sparse_kantorovich,
-    "wasserstein": sparse_kantorovich,
     "braycurtis": sparse_bray_curtis,
     # Binary distances
     "hamming": sparse_hamming,
@@ -837,7 +1077,19 @@ sparse_named_distances = {
     # Angular distances
     "cosine": sparse_cosine,
     "correlation": sparse_correlation,
+    # Distribution distances
+    "kantorovich": sparse_kantorovich,
+    "wasserstein": sparse_kantorovich,
+    "wasserstein_1d": sparse_wasserstein_1d,
+    "wasserstein-1d": sparse_wasserstein_1d,
+    "kantorovich-1d": sparse_wasserstein_1d,
+    "kantorovich-1d": sparse_wasserstein_1d,
     "hellinger": sparse_hellinger,
+    "jensen-shannon": sparse_jensen_shannon_divergence,
+    "jensen_shannon": sparse_jensen_shannon_divergence,
+    "symmetric-kl": sparse_symmetric_kl_divergence,
+    "symmetric_kl": sparse_symmetric_kl_divergence,
+    "symmetric_kullback_liebler": sparse_symmetric_kl_divergence,
 }
 
 sparse_need_n_features = (
