@@ -6,17 +6,17 @@ import time
 
 import numba
 from numba.core import types
-from numba.experimental import structref
+import numba.experimental.structref as structref
 import numpy as np
 
 
-@numba.njit("void(i8[:], i8)")
+@numba.njit("void(i8[:], i8)", cache=True)
 def seed(rng_state, seed):
     """Seed the random number generator with a given seed."""
     rng_state.fill(seed + 0xFFFF)
 
 
-@numba.njit("i4(i8[:])")
+@numba.njit("i4(i8[:])", cache=True)
 def tau_rand_int(state):
     """A fast (pseudo)-random number generator.
 
@@ -42,7 +42,7 @@ def tau_rand_int(state):
     return state[0] ^ state[1] ^ state[2]
 
 
-@numba.njit("f4(i8[:])")
+@numba.njit("f4(i8[:])", cache=True)
 def tau_rand(state):
     """A fast (pseudo)-random number generator for floats in the range [0,1]
 
@@ -69,9 +69,10 @@ def tau_rand(state):
     locals={
         "dim": numba.types.intp,
         "i": numba.types.uint32,
-        "result": numba.types.float32,
+        # "result": numba.types.float32, # This provides speed, but causes errors in corner cases
     },
     fastmath=True,
+    cache=True,
 )
 def norm(vec):
     """Compute the (standard l2) norm of a vector.
@@ -91,7 +92,7 @@ def norm(vec):
     return np.sqrt(result)
 
 
-@numba.njit()
+@numba.njit(cache=True)
 def rejection_sample(n_samples, pool_size, rng_state):
     """Generate n_samples many integers from 0 to pool_size such that no
     integer is selected twice. The duplication constraint is achieved via
@@ -147,31 +148,27 @@ class Heap(structref.StructRefProxy):
         return Heap_get_flags(self)
 
 
-@numba.njit
+@numba.njit(cache=True)
 def Heap_get_flags(self):
     return self.flags
 
 
-@numba.njit
+@numba.njit(cache=True)
 def Heap_get_distances(self):
     return self.distances
 
 
-@numba.njit
+@numba.njit(cache=True)
 def Heap_get_indices(self):
     return self.indices
 
 
-structref.define_proxy(
-    Heap,
-    HeapType,
-    ["indices", "distances", "flags"],
-)
+structref.define_proxy(Heap, HeapType, ["indices", "distances", "flags"])
 
 # Heap = namedtuple("Heap", ("indices", "distances", "flags"))
 
 
-@numba.njit()
+@numba.njit(cache=True)
 def make_heap(n_points, size):
     """Constructor for the numba enabled heap objects. The heaps are used
     for approximate nearest neighbor search, maintaining a list of potential
@@ -202,197 +199,7 @@ def make_heap(n_points, size):
     return result
 
 
-@numba.jit(
-    locals={
-        "indices": numba.types.int32[::1],
-        "weights": numba.types.float32[::1],
-        "is_new": numba.types.uint8[::1],
-        "i": numba.types.uint16,
-        "ic1": numba.types.uint16,
-        "ic2": numba.types.uint16,
-        "i_swap": numba.types.uint16,
-        "heap_size": numba.types.uint16,
-    }
-)
-def heap_push(heap, row, weight, index, flag):
-    """Push a new element onto the heap. The heap stores potential neighbors
-    for each graph_data point. The ``row`` parameter determines which graph_data point we
-    are addressing, the ``weight`` determines the distance (for heap sorting),
-    the ``index`` is the element to add, and the flag determines whether this
-    is to be considered a new addition.
-
-    Parameters
-    ----------
-    heap: ndarray generated by ``make_heap``
-        The heap object to push into
-
-    row: int
-        Which actual heap within the heap object to push to
-
-    weight: float
-        The priority value of the element to push onto the heap
-
-    index: int
-        The actual value to be pushed
-
-    flag: int
-        Whether to flag the newly added element or not.
-
-    Returns
-    -------
-    success: The number of new elements successfully pushed into the heap.
-    """
-    row = np.int32(row)
-    weight = np.float32(weight)
-    index = np.int32(index)
-    flag = np.uint8(flag)
-
-    indices = heap[0][row]
-    weights = heap[1][row]
-    is_new = heap[2][row]
-
-    if weight >= weights[0]:
-        return 0
-
-    # break if we already have this element.
-    for i in range(indices.shape[0]):
-        if index == indices[i]:
-            return 0
-
-    # insert val at position zero
-    weights[0] = weight
-    indices[0] = index
-    is_new[0] = flag
-
-    # descend the heap, swapping values until the max heap criterion is met
-    i = 0
-    while True:
-        ic1 = 2 * i + 1
-        ic2 = ic1 + 1
-
-        if ic1 >= indices.shape[0]:
-            break
-        elif ic2 >= indices.shape[0]:
-            if weights[ic1] > weight:
-                i_swap = ic1
-            else:
-                break
-        elif weights[ic1] >= weights[ic2]:
-            if weight < weights[ic1]:
-                i_swap = ic1
-            else:
-                break
-        else:
-            if weight < weights[ic2]:
-                i_swap = ic2
-            else:
-                break
-
-        weights[i] = weights[i_swap]
-        indices[i] = indices[i_swap]
-        is_new[i] = is_new[i_swap]
-
-        i = i_swap
-
-    weights[i] = weight
-    indices[i] = index
-    is_new[i] = flag
-
-    return 1
-
-
-@numba.jit(
-    locals={
-        "indices": numba.types.int32[::1],
-        "weights": numba.types.float32[::1],
-        "is_new": numba.types.uint8[::1],
-        "i": numba.types.uint16,
-        "ic1": numba.types.uint16,
-        "ic2": numba.types.uint16,
-        "i_swap": numba.types.uint16,
-        "heap_size": numba.types.uint16,
-    }
-)
-def unchecked_heap_push(heap, row, weight, index, flag):
-    """Push a new element onto the heap. The heap stores potential neighbors
-    for each graph_data point. The ``row`` parameter determines which graph_data point we
-    are addressing, the ``weight`` determines the distance (for heap sorting),
-    the ``index`` is the element to add, and the flag determines whether this
-    is to be considered a new addition.
-
-    Parameters
-    ----------
-    heap: ndarray generated by ``make_heap``
-        The heap object to push into
-
-    row: int
-        Which actual heap within the heap object to push to
-
-    weight: float
-        The priority value of the element to push onto the heap
-
-    index: int
-        The actual value to be pushed
-
-    flag: int
-        Whether to flag the newly added element or not.
-
-    Returns
-    -------
-    success: The number of new elements successfully pushed into the heap.
-    """
-    if weight >= heap[1][row, 0]:
-        return 0
-
-    indices = heap[0][row]
-    weights = heap[1][row]
-    is_new = heap[2][row]
-
-    # insert val at position zero
-    weights[0] = weight
-    indices[0] = index
-    is_new[0] = flag
-
-    heap_size = indices.shape[0]
-
-    # descend the heap, swapping values until the max heap criterion is met
-    i = 0
-    while True:
-        ic1 = 2 * i + 1
-        ic2 = ic1 + 1
-
-        if ic1 >= heap_size:
-            break
-        elif ic2 >= heap_size:
-            if weights[ic1] > weight:
-                i_swap = ic1
-            else:
-                break
-        elif weights[ic1] >= weights[ic2]:
-            if weight < weights[ic1]:
-                i_swap = ic1
-            else:
-                break
-        else:
-            if weight < weights[ic2]:
-                i_swap = ic2
-            else:
-                break
-
-        weights[i] = weights[i_swap]
-        indices[i] = indices[i_swap]
-        is_new[i] = is_new[i_swap]
-
-        i = i_swap
-
-    weights[i] = weight
-    indices[i] = index
-    is_new[i] = flag
-
-    return 1
-
-
-@numba.njit()
+@numba.njit(cache=True)
 def siftdown(heap1, heap2, elt):
     """Restore the heap property for a heap with an out of place element
     at position ``elt``. This works with a heap pair where heap1 carries
@@ -416,95 +223,79 @@ def siftdown(heap1, heap2, elt):
             elt = swap
 
 
-@numba.njit()
-def deheap_sort(heap):
-    """Given an array of heaps (of graph_indices and weights), unpack the heap
-    out to give and array of sorted lists of graph_indices and weights by increasing
-    weight. This is effectively just the second half of heap sort (the first
-    half not being required since we already have the graph_data in a heap).
+@numba.njit(parallel=True, cache=False)
+def deheap_sort(indices, distances):
+    """Given two arrays representing a heap (indices and distances), reorder the 
+     arrays by increasing distance. This is effectively just the second half of
+     heap sort (the first half not being required since we already have the
+     graph_data in a heap).
+
+     Note that this is done in-place.
 
     Parameters
     ----------
-    heap : array of shape (3, n_samples, n_neighbors)
-        The heap to turn into sorted lists.
+    indices : array of shape (n_samples, n_neighbors)
+        The graph indices to sort by distance.
+    distances : array of shape (n_samples, n_neighbors)
+        The corresponding edge distance.
 
     Returns
     -------
-    graph_indices, weights: arrays of shape (n_samples, n_neighbors)
-        The graph_indices and weights sorted by increasing weight.
+    indices, distances: arrays of shape (n_samples, n_neighbors)
+        The indices and distances sorted by increasing distance.
     """
-    indices = heap[0]
-    weights = heap[1]
+    for i in numba.prange(indices.shape[0]):
+        # starting from the end of the array and moving back
+        for j in range(indices.shape[1] - 1, 0, -1):
+            indices[i, 0], indices[i, j] = indices[i, j], indices[i, 0]
+            distances[i, 0], distances[i, j] = distances[i, j], distances[i, 0]
 
-    for i in range(indices.shape[0]):
+            siftdown(distances[i, :j], indices[i, :j], 0)
 
-        ind_heap = indices[i]
-        dist_heap = weights[i]
-
-        for j in range(ind_heap.shape[0] - 1):
-            ind_heap[0], ind_heap[ind_heap.shape[0] - j - 1] = (
-                ind_heap[ind_heap.shape[0] - j - 1],
-                ind_heap[0],
-            )
-            dist_heap[0], dist_heap[dist_heap.shape[0] - j - 1] = (
-                dist_heap[dist_heap.shape[0] - j - 1],
-                dist_heap[0],
-            )
-
-            siftdown(
-                dist_heap[: dist_heap.shape[0] - j - 1],
-                ind_heap[: ind_heap.shape[0] - j - 1],
-                0,
-            )
-
-    return indices.astype(np.int64), weights
+    return indices, distances
 
 
-@numba.njit()
-def smallest_flagged(heap, row):
-    """Search the heap for the smallest element that is
-    still flagged.
-
-    Parameters
-    ----------
-    heap: array of shape (3, n_samples, n_neighbors)
-        The heaps to search
-
-    row: int
-        Which of the heaps to search
-
-    Returns
-    -------
-    index: int
-        The index of the smallest flagged element
-        of the ``row``th heap, or -1 if no flagged
-        elements remain in the heap.
-    """
-    ind = heap[0][row]
-    dist = heap[1][row]
-    flag = heap[2][row]
-
-    min_dist = np.inf
-    result_index = -1
-
-    for i in range(ind.shape[0]):
-        if flag[i] == 1 and dist[i] < min_dist:
-            min_dist = dist[i]
-            result_index = i
-
-    if result_index >= 0:
-        flag[result_index] = 0.0
-        return int(ind[result_index])
-    else:
-        return -1
+# @numba.njit()
+# def smallest_flagged(heap, row):
+#     """Search the heap for the smallest element that is
+#     still flagged.
+#
+#     Parameters
+#     ----------
+#     heap: array of shape (3, n_samples, n_neighbors)
+#         The heaps to search
+#
+#     row: int
+#         Which of the heaps to search
+#
+#     Returns
+#     -------
+#     index: int
+#         The index of the smallest flagged element
+#         of the ``row``th heap, or -1 if no flagged
+#         elements remain in the heap.
+#     """
+#     ind = heap[0][row]
+#     dist = heap[1][row]
+#     flag = heap[2][row]
+#
+#     min_dist = np.inf
+#     result_index = -1
+#
+#     for i in range(ind.shape[0]):
+#         if flag[i] == 1 and dist[i] < min_dist:
+#             min_dist = dist[i]
+#             result_index = i
+#
+#     if result_index >= 0:
+#         flag[result_index] = 0.0
+#         return int(ind[result_index])
+#     else:
+#         return -1
 
 
-@numba.njit(parallel=True, locals={"idx": numba.types.int64})
-def new_build_candidates(
-    current_graph,
-    max_candidates,
-    rng_state,
-):
+@numba.njit(parallel=True, locals={"idx": numba.types.int64}, cache=False)
+def new_build_candidates(current_graph, max_candidates, rng_state, n_threads):
     """Build a heap of candidate neighbors for nearest neighbor descent. For
     each vertex the candidate neighbors are any current neighbors, and any
     vertices that have the vertex as one of their nearest neighbors.
@@ -541,8 +332,6 @@ def new_build_candidates(
         (n_vertices, max_candidates), np.inf, dtype=np.float32
     )
 
-    n_threads = numba.get_num_threads()
-
     for n in numba.prange(n_threads):
         local_rng_state = rng_state + n
         for i in range(n_vertices):
@@ -558,10 +347,7 @@ def new_build_candidates(
                 if isn:
                     if i % n_threads == n:
                         checked_heap_push(
-                            new_candidate_priority[i],
-                            new_candidate_indices[i],
-                            d,
-                            idx,
+                            new_candidate_priority[i], new_candidate_indices[i], d, idx
                         )
                     if idx % n_threads == n:
                         checked_heap_push(
@@ -573,10 +359,7 @@ def new_build_candidates(
                 else:
                     if i % n_threads == n:
                         checked_heap_push(
-                            old_candidate_priority[i],
-                            old_candidate_indices[i],
-                            d,
-                            idx,
+                            old_candidate_priority[i], old_candidate_indices[i], d, idx
                         )
                     if idx % n_threads == n:
                         checked_heap_push(
@@ -601,14 +384,14 @@ def new_build_candidates(
     return new_candidate_indices, old_candidate_indices
 
 
-@numba.njit("b1(u1[::1],i4)")
+@numba.njit("b1(u1[::1],i4)", cache=True)
 def has_been_visited(table, candidate):
     loc = candidate >> 3
     mask = 1 << (candidate & 7)
     return table[loc] & mask
 
 
-@numba.njit("void(u1[::1],i4)")
+@numba.njit("void(u1[::1],i4)", cache=True)
 def mark_visited(table, candidate):
     loc = candidate >> 3
     mask = 1 << (candidate & 7)
@@ -626,6 +409,7 @@ def mark_visited(table, candidate):
         "ic2": numba.types.uint16,
         "i_swap": numba.types.uint16,
     },
+    cache=True,
 )
 def simple_heap_push(priorities, indices, p, n):
     if p >= priorities[0]:
@@ -682,6 +466,7 @@ def simple_heap_push(priorities, indices, p, n):
         "ic2": numba.types.uint16,
         "i_swap": numba.types.uint16,
     },
+    cache=True,
 )
 def checked_heap_push(priorities, indices, p, n):
     if p >= priorities[0]:
@@ -743,65 +528,7 @@ def checked_heap_push(priorities, indices, p, n):
         "ic2": numba.types.uint16,
         "i_swap": numba.types.uint16,
     },
-)
-def flagged_heap_push(priorities, indices, flags, p, n, f):
-    if p >= priorities[0]:
-        return 0
-
-    size = priorities.shape[0]
-
-    # insert val at position zero
-    priorities[0] = p
-    indices[0] = n
-    flags[0] = f
-
-    # descend the heap, swapping values until the max heap criterion is met
-    i = 0
-    while True:
-        ic1 = 2 * i + 1
-        ic2 = ic1 + 1
-
-        if ic1 >= size:
-            break
-        elif ic2 >= size:
-            if priorities[ic1] > p:
-                i_swap = ic1
-            else:
-                break
-        elif priorities[ic1] >= priorities[ic2]:
-            if p < priorities[ic1]:
-                i_swap = ic1
-            else:
-                break
-        else:
-            if p < priorities[ic2]:
-                i_swap = ic2
-            else:
-                break
-
-        priorities[i] = priorities[i_swap]
-        indices[i] = indices[i_swap]
-        flags[i] = flags[i_swap]
-
-        i = i_swap
-
-    priorities[i] = p
-    indices[i] = n
-    flags[i] = f
-
-    return 1
-
-
-@numba.njit(
-    "i4(f4[::1],i4[::1],u1[::1],f4,i4,u1)",
-    fastmath=True,
-    locals={
-        "size": numba.types.intp,
-        "i": numba.types.uint16,
-        "ic1": numba.types.uint16,
-        "ic2": numba.types.uint16,
-        "i_swap": numba.types.uint16,
-    },
+    cache=True,
 )
 def checked_flagged_heap_push(priorities, indices, flags, p, n, f):
     if p >= priorities[0]:
@@ -867,14 +594,15 @@ def checked_flagged_heap_push(priorities, indices, flags, p, n, f):
         "i": numba.uint32,
         "j": numba.uint32,
     },
+    cache=False,
 )
-def apply_graph_updates_low_memory(current_graph, updates):
+def apply_graph_updates_low_memory(current_graph, updates, n_threads):
 
     n_changes = 0
     priorities = current_graph[1]
     indices = current_graph[0]
     flags = current_graph[2]
-    n_threads = numba.get_num_threads()
+    # n_threads = numba.get_num_threads()
 
     for n in numba.prange(n_threads):
         for i in range(len(updates)):
@@ -885,33 +613,21 @@ def apply_graph_updates_low_memory(current_graph, updates):
                     continue
 
                 if p % n_threads == n:
-                    # added = heap_push(current_graph, p, d, q, 1)
                     added = checked_flagged_heap_push(
-                        priorities[p],
-                        indices[p],
-                        flags[p],
-                        d,
-                        q,
-                        1,
+                        priorities[p], indices[p], flags[p], d, q, 1
                     )
                     n_changes += added
 
                 if q % n_threads == n:
-                    # added = heap_push(current_graph, q, d, p, 1)
                     added = checked_flagged_heap_push(
-                        priorities[q],
-                        indices[q],
-                        flags[q],
-                        d,
-                        p,
-                        1,
+                        priorities[q], indices[q], flags[q], d, p, 1
                     )
                     n_changes += added
 
     return n_changes
 
 
-@numba.njit(locals={"p": numba.types.int64, "q": numba.types.int64})
+@numba.njit(locals={"p": numba.types.int64, "q": numba.types.int64}, cache=True)
 def apply_graph_updates_high_memory(current_graph, updates, in_graph):
 
     n_changes = 0
@@ -928,8 +644,7 @@ def apply_graph_updates_high_memory(current_graph, updates, in_graph):
             elif q in in_graph[p]:
                 pass
             else:
-                # added = unchecked_heap_push(current_graph, p, d, q, 1)
-                added = flagged_heap_push(
+                added = checked_flagged_heap_push(
                     current_graph[1][p],
                     current_graph[0][p],
                     current_graph[2][p],
@@ -945,8 +660,7 @@ def apply_graph_updates_high_memory(current_graph, updates, in_graph):
             if p == q or p in in_graph[q]:
                 pass
             else:
-                # added = unchecked_heap_push(current_graph, q, d, p, 1)
-                added = flagged_heap_push(
+                added = checked_flagged_heap_push(
                     current_graph[1][p],
                     current_graph[0][p],
                     current_graph[2][p],
@@ -962,7 +676,7 @@ def apply_graph_updates_high_memory(current_graph, updates, in_graph):
     return n_changes
 
 
-@numba.njit()
+@numba.njit(cache=True)
 def initalize_heap_from_graph_indices(heap, graph_indices, data, metric):
 
     for i in range(graph_indices.shape[0]):
@@ -970,12 +684,12 @@ def initalize_heap_from_graph_indices(heap, graph_indices, data, metric):
             j = graph_indices[i, idx]
             if j >= 0:
                 d = metric(data[i], data[j])
-                flagged_heap_push(heap[1][i], heap[0][i], heap[2][i], d, j, 1)
+                checked_flagged_heap_push(heap[1][i], heap[0][i], heap[2][i], d, j, 1)
 
     return heap
 
 
-@numba.njit(parallel=True)
+@numba.njit(parallel=True, cache=False)
 def sparse_initalize_heap_from_graph_indices(
     heap, graph_indices, data_indptr, data_indices, data_vals, metric
 ):
@@ -988,8 +702,7 @@ def sparse_initalize_heap_from_graph_indices(
             ind2 = data_indices[data_indptr[j] : data_indptr[j + 1]]
             data2 = data_vals[data_indptr[j] : data_indptr[j + 1]]
             d = metric(ind1, data1, ind2, data2)
-            # unchecked_heap_push(heap, i, d, j, 1)
-            flagged_heap_push(heap[0][i], heap[1][i], heap[2][i], j, d, 1)
+            checked_flagged_heap_push(heap[1][i], heap[0][i], heap[2][i], d, j, 1)
 
     return heap
 
